
(module mreddesigner-function mzscheme

  (provide get-property-value
           get-property-children
           ;;;
           set-property-value
           set-property-children
           set-property-parent
           set-property-select
           set-property-collapse
           set-property-pointer
           set-property-double-value
           ;;;
           apply-function-on-data-list
           apply-procedure-on-data-list
           apply-function-on-data-list-improved
  )
  
  ;; Properties
  ;; ---------------------------------------------------------------------------

  ;; This function returns the property value, given a line
  (define (get-property-value line property)
    (if (assq property (cdr line))
      (cadr (assq property (cdr line)))
      #f
    )
  )
  
  ;; This function gets the children of the given line
  (define (get-property-children data-list)
    (if (assq 'children (cdr data-list))
      (cdr (assq 'children (cdr data-list)))
      #f))
  

  ;; data-list: the list containing the properties
  ;;            first element is a symbol
  ;;            the rest are lists, key + value
  ;; prop: symbol to identify the property, it will be compared with eq?
  ;; func-val: a function with one argument, 
  ;;           The argument for the function will be the current value of the property
  ;;           as a list.
  ;;           If there was no value for the property the argument is the no-prop symbol.
  ;;           The return value of the function will be assigned to the property.
  ;;           If the returned value is an atom then it performs (list prop returned)
  ;;           If the returned value is a list then it performs (cons prop returned)
  (define (set-property-value data-list prop func-val)
    (cond
      ; could not find the property
      ((null? data-list) 
       (let
         ; no property, so pass the no-prop symbol to the function
         ((val (func-val 'no-prop)))
         (list (if (list? val) (cons prop val) (list prop val)))
       )
      )
      ; it is not a property, just add it to the list
      ((symbol? (car data-list))
       (cons (car data-list) 
             (set-property-value (cdr data-list) prop func-val))
      )
      ; if the front element is a list with the property
      ((eq? (caar data-list) prop)
       (let*
         ; get the rest of the property
         ((rest (cdar data-list))
          ; if property list is empty pass the no-prop symbol to the function
          (val  (func-val (if (null? rest) 'no-prop rest)))
         )
         ; just add it to the list, no further recursion
         (cons (if (list? val) (cons prop val) (list prop val))
               (cdr data-list))
       )
      )
      (else 
       (cons (car data-list) 
             (set-property-value (cdr data-list) prop func-val))
      )
    )
  )
  
  ;; This sub-routine sets a children list to a new list
  (define (set-property-children data-list children)
    (set-property-value data-list 'children (lambda (x) children))
  )
  
  ; Replaces the current parent with a new parent
  (define (set-property-parent data-list parent)
    (set-property-value data-list 'parent (lambda (x) parent))
  )
  
  ;; This sub-routine selects/deselects an element
  (define (set-property-select data-list selected)
    (set-property-value data-list 'selected (lambda (x) selected))
  )  
  
  ;; This sub-routine collapses/uncollapses an element
  ;; It inverts the state
  (define (set-property-collapse data-list change)
    (set-property-value data-list 'collapsed 
                        (lambda (x) (if change (not (car x)) (car x))))
  )
  
  ; Replaces the current pointer with a new pointer
  (define (set-property-pointer data-list pointer)
    (set-property-value data-list 'pointer (lambda (x) pointer))
  )

  
  ;; This is a special function it overwrites two properties at the same time
  ;; for example x and y position or width and height.
  ;; The two properties should not be the same. The first property is checked first.
  ;; If the properties are not in the list they are added, and only the one is 
  ;; added which is not in the list.
  (define (set-property-double-value data-list p1 v1 p2 v2)
    (cond
      ; could not find the properties
      ((null? data-list)
       (append (if p1 (list (list p1 v1)) '())
               (if p2 (list (list p2 v2)) '())
               data-list)
      )
      ; it is not a property, just add it to the list
      ((symbol? (car data-list))
       (cons (car data-list) 
             (set-property-double-value (cdr data-list) p1 v1 p2 v2))
      )
      ; if the front element is a list with the property 1
      ((eq? (caar data-list) p1)
       (cons (list p1 v1) 
             (set-property-double-value (cdr data-list) #f #f p2 v2))
      )
      ; if the front element is a list with the property 2
      ((eq? (caar data-list) p2)
       (cons (list p2 v2) 
             (set-property-double-value (cdr data-list) p1 v1 #f #f))
      )
      (else 
       (cons (car data-list) 
             (set-property-double-value (cdr data-list) p1 v1 p2 v2))
      )
    )
  )
  
  
  
  ;; Functions to operate on the whole lists
  ;; ---------------------------------------------------------------------------
  
  ;; This function applies a function to each line and returns the
  ;; built data-list.
  (define (apply-function-on-data-list data-list function)
    ;; The helper function
    (define (helper data-list)
      (if (null? data-list)
        data-list
        (begin
          ;; Apply the right function
          (let* ((mod-current-line (function (car data-list)))
                 ;; First recursive call (for children)
                 (mod-current-children (if (assq 'children (cdar data-list)) 
                                         (helper (cdr (assq 'children (cdr mod-current-line))))
                                         #f)))
            ;; We build the result and call the second recursive call (for brothers)
            (cons (if mod-current-children
                      (set-property-children mod-current-line mod-current-children)
                      mod-current-line)
                  (helper (cdr data-list)))))))
    ;; The function body
    (helper data-list))

  ;; This function applies a procedure to each line and returns void.
  (define (apply-procedure-on-data-list data-list function)
    ;; The helper function
    (define (helper data-list)
      (if (null? data-list)
        (void)
        (begin
          ;; Apply the right function
          (function (car data-list))
          ;; First recursive call (for children)
          (if (assq 'children (cdar data-list)) 
              (helper (cdr (assq 'children (cdar data-list)))))
          ;; Second recursive call (for brothers)
          (helper (cdr data-list)))))
    ;; The function body
    (helper data-list))

  ;; This is an improved version, which allows to delete lines.
(define (apply-function-on-data-list-improved data-list function)
  (define (set-children data-list children)
    (cond ((null? data-list) 
           data-list)
          ((symbol? (car data-list)) 
           (cons (car data-list) (set-children (cdr data-list) children)))
          ((and (eq? (caar data-list) 'children) 
                (null? children) 
                (set-children (cdr data-list) children)))
          ((eq? (caar data-list) 'children) 
           (cons (cons 'children children) (set-children (cdr data-list) children)))
          (else (cons (car data-list) (set-children (cdr data-list) children)))))
  ;; The helper function
  (define (helper data-list)
    (if (null? data-list)
        data-list
        (begin
          ;; Apply the right function
          (let* ((mod-current-line (function (car data-list)))
                 ;; First recursive call (for children)
                 (mod-current-children (if (not (eq? mod-current-line 'nil))
                                           (if (assq 'children (cdar data-list))
                                               (helper (cdr (assq 'children (cdr mod-current-line))))
                                               #f)
                                           #f)))
            ;; We build the result and call the second recursive call (for brothers)
            (if (eq? mod-current-line 'nil)
                (helper (cdr data-list))
                (cons (if mod-current-children
                          (set-children mod-current-line mod-current-children)
                          mod-current-line)
                      (helper (cdr data-list))))))))
  ;; The function body
  (helper data-list))

  
)

