;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - preview.ss                                                     #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

;; This file contains general functions used in previewing.


; This is a special class for previewing frames (windows).
; It registers its position and its size in the data structure,
; so it is possible to restore.
(define preview-frame%
  (class frame%
    ; id of the frame in the data list
    (init-field (mred-id #f))
    (define/public (mred-id-set! id)
      (set! mred-id id))
    
    ; registering functions
    (override on-move on-size)
    (define (on-move x y)
      (let*
        ((pos-func (lambda (line)
                     (if (eq? (get-property-value line 'id) mred-id)
                       (set-property-double-value line 'x x 'y y)
                       line))))
        ; change the global data structure
        (set! project-data-global
              (apply-function-on-data-list project-data-global pos-func))
        ; notify the hierarchy window
        (send hierarchy-widget set-data project-data-global)
        ; notify the property window
        (send properties-widget update-in)
      )
    )
    
    (define (on-size w h)
      (let*
        ((pos-func (lambda (line)
                     (if (eq? (get-property-value line 'id) mred-id)
                       (set-property-double-value line 'width w 'height h)
                       line))))
        ; change the global data structure
        (set! project-data-global
              (apply-function-on-data-list project-data-global pos-func))
        ; notify the hierarchy window
        (send hierarchy-widget set-data project-data-global)
        ; notify the property window
        (send properties-widget update-in)
      )
    )

    (super-new)
  )
)


;; This function removes the properties which cannot be interpreted by MrEd.
(define (del-property-no-mred line parent)
  (cond ((null? line) line)
        ;; We don't want symbols
        ((symbol? (car line)) (del-property-no-mred (cdr line) parent))
        ;; All these tags are not properties
        ((eq? (caar line) 'children) (del-property-no-mred (cdr line) parent))
        ((eq? (caar line) 'necessary) (del-property-no-mred (cdr line) parent))
        ((eq? (caar line) 'optional) (del-property-no-mred (cdr line) parent))
        ((eq? (caar line) 'selected) (del-property-no-mred (cdr line) parent))
        ((eq? (caar line) 'collapsed) (del-property-no-mred (cdr line) parent))
        ((eq? (caar line) 'id) (del-property-no-mred (cdr line) parent))
        ;; This one was added to be able to reload the project "on the fly".
        ((eq? (caar line) 'pointer) (del-property-no-mred (cdr line) parent))
        ;; We need to update the "parent" property
        ((eq? (caar line) 'parent) 
         (cons (list 'parent parent) (del-property-no-mred (cdr line) parent)))
        ;; If this is a valid property, continue with the remaining list...
        (else (cons (car line) (del-property-no-mred (cdr line) parent)))))

;; This function creates a frame (window) for preview
;; It uses a special class, which can store its position and size in the
;; global data structure
(define (preview-create-frame line parent)
  (eval (append '(new preview-frame%)
                (list (list 'mred-id 
                            (list 'quote 
                                  (get-property-value line 'id))))
                (del-property-no-mred line parent)))
)

;; This function creates a MrEd widget for preview
(define (preview-create-widget line parent)
  (eval (append (list 'new) 
                (list (string->symbol 
                       (string-append 
                        (symbol->string (car line)) "%")))
                (del-property-no-mred line parent)))
)


;; Deletes an element in the preview
(define (preview-delete-element element)
  ;; Is the element a frame?
  (if (eq? (car element) 'frame)
      ;; Then we have no choice, then - just hide it.
      (send (get-property-value element 'pointer) show #f)
      ;; Otherwise, we need to check whether we can update the preview...
      (if (member (car element) '(pane horizontal-pane vertical-pane))
          ;; We can't without reloading the whole project!
          (preview-update-preview project-data-global project-data-global)
          ;; Otherwise, we get a pointer to the parent...
          (send (get-property-value (project-data-get-parent element) 'pointer)
                ;; To which we ask to delete a child...
                delete-child
                ;; Whose pointer is obtained with the element line...
                (get-property-value element 'pointer)))))

;; This function is called for example when the user opens a new file : it
;; then removes the preview from the previous file and creates the preview
;; from the new file...
;; It is also called when the user updates the preview "on the fly" (when
;; the user changes the style property of a container).
(define (preview-update-preview old-data new-data)
  ;; A function which allows us to simply hide frames.
  ;; (There is no way to delete them.)
  (define (hide-if-frame line)
    (if (and (eq? (car line) 'frame)
             (get-property-value line 'pointer))
      (send (get-property-value line 'pointer) show #f)))
  ; this function can be used to display all windows or frames
  (define (show-if-frame line)
    (if (eq? (car line) 'frame)
      (send (get-property-value line 'pointer) show #t)))
  ;; Then we apply the function on each element of the list.
  (apply-procedure-on-data-list old-data hide-if-frame)
  
  ;; Ok, we have hidden all the widgets from the old-data. Now we are going
  ;; to parse the new list, and for each line, we will first create the associated
  ;; widget, and then update the line so that the pointer to the widget is inserted
  ;; into it. This might be the most difficult part of this project - but I will try
  ;; to comment it as much as possible.
  
  ;; First we create a parser that will take a line from the new data, and computes it
  ;; so that it can be passed to new, in order to create the widget
  (letrec 
    (;; Then we develop a function that will take a line from the new data, 
     ;; call the necessary functions to preview it,
     ;; then return two results : the computed line and the pointer to the newly created widget.
     (line-helper 
      (lambda (line parent)
        ;; We ignore the "Project" line...
        (cond ((eq? (car line) 'project) (cons line #f))
              ;; For each line, we call 'new' with the right arguments
              (else
               (let*
                 ((pointer 
                   (if (eq? (car line) 'frame)
                     ; this is a special widget
                     (preview-create-frame line parent)
                     ; this is normal widget
                     (preview-create-widget line parent))))
                 ;; And then, we send the two needed results.
                 (cons (set-property-pointer line pointer) pointer))))))
          
     ;; Then, the main loop, which computes the whole data list - 
     ;; mainly inspired from the apply-function-on-data-list function,
     ;; but slightly more complex, because we have to deal with the widgets' parents.
     (list-helper 
      (lambda (data-list parent)
        ;; If the data-list is empty, the job is done!
        (if (null? data-list)
          data-list
          ;; Otherwise, we apply line-helper on the line...
          (let* 
            ((results (line-helper (car data-list) parent))
             ;; And dispatch the results...
             (mod-current-ptr (cdr results))
             (mod-current-line (car results))
             
             ;; First recursive call (for children)... 
             ;; To which we pass mod-current-ptr because
             ;; the pointer to their parent is the pointer to the current element...
             (mod-current-children 
              (if (assq 'children (cdar data-list))
                (list-helper (cdr (assq 'children (cdr mod-current-line)))
                             mod-current-ptr) 
                #f
              )
             )
            )
            
            ;; We build the result and call the second recursive call (for brothers)... 
            ;; To which we pass the current parent because our brother and 
            ;; us have the same parents!
            (cons (if mod-current-children
                    (set-property-children mod-current-line mod-current-children)
                    mod-current-line)
                  (list-helper (cdr data-list) parent)))))))
    
    ;; And then we call list-helper, and update project-data-global with the result...
    (set! project-data-global (list-helper new-data #f))
  )
  
  ; Finally show the windows
  (apply-procedure-on-data-list project-data-global show-if-frame)
)


;; This function updates, in the preview, the order of the element's children
(define (preview-update-children-order element)
  ;; This filter returns the updated children list...
  (define (widget-filter widget-list)
    ;; Just get the children list and compute it into a list of pointers...
    (map (lambda (elt) (get-property-value elt 'pointer)) 
         (filter (lambda (elt) (not (eq? (car elt) 'frame))) 
                 (get-property-children element))))
  ;; We ask the widget to update its children using the filter we just wrote...
  (send (get-property-value element 'pointer) change-children widget-filter)
)


;; This function is called in the properties-widget on update-out.
(define (preview-update-widget widget-line convert-elt last-value-changed) 
  ;; If the widget is a container and the last value changed is a property that we
  ;; cannot change "on the fly", then, we build the whole widget again!
  (if (and (not (container? widget-line))
           (or (eq? 'choices last-value-changed)
               (eq? 'label last-value-changed)
               (eq? 'callback last-value-changed)
               (eq? 'style last-value-changed)
               (eq? 'min-value last-value-changed)
               (eq? 'max-value last-value-changed)))
      ;; We will have to save a pointer to the newly created element...
      (let ((new-pointer '?))
        ;; This parser allows us to compute the line in a call to "new"...
        (define (parse-list L)
          (cond ((null? L) '())
                ((eq? (caar L) 'parent) 
                 (cons (list 'parent 
                             (get-property-value (project-data-get-parent widget-line) 'pointer))
                       (parse-list (cdr L))))
                ((eq? (caar L) 'necessary)(parse-list (cdr L)))
                ((eq? (caar L) 'optional)(parse-list (cdr L)))
                ((eq? (caar L) 'selected)(parse-list (cdr L)))
                ((eq? (caar L) 'collapsed)(parse-list (cdr L)))
                ((eq? (caar L) 'pointer)(parse-list (cdr L)))
                ((eq? (caar L) 'children) (parse-list (cdr L)))
                ((eq? (caar L) 'id)(parse-list (cdr L)))
                ((eq? (cadar L) 'na) (parse-list (cdr L)))
                (else (cons (car L) (parse-list (cdr L))))))
        ;; This function parses the parent's children list... Looking for the current widget.
        (define (filter widget-list)
          ;; If the list is empty, then we have nothing to do...
          (cond ((null? widget-list) widget-list)
                ;; Did we find the current element?
                ((eq? (car widget-list) (get-property-value widget-line 'pointer))
                 ;; Yes we did - therefore we instantiate it again...
                 (let ((new-element (eval `(new ,(symbol-append (car widget-line) '%)
                                                ,@(parse-list (cdr widget-line))))))
                   ;; ...save the pointer...
                   (set! new-pointer new-element)
                   ;; ...and return the updated children list.
                   (cons new-element (cdr widget-list))))
                ;; Otherwise, we do not change the currently parsed element.
                (else (cons (car widget-list) (filter (cdr widget-list))))))
        
        ;; This function updates the current widget line, 
        ;; setting the right pointer for the pointer property.
        (define (update-line line)
          ;; If the line is empty, do nothing.
          (cond ((null? line) line)
                ;; If the first element is a symbol, just keep it.
                ((symbol? (car line))(cons (car line) (update-line (cdr line))))
                ;; If the first element is the pointer, update it.
                ((eq? (caar line) 'pointer) 
                 (cons (list 'pointer new-pointer) (update-line (cdr line))))
                ;; Otherwise, just keep the first element.
                (else (cons (car line) (update-line (cdr line))))))
        
        ;; We ask the widget's parent...
        (send (get-property-value (project-data-get-parent widget-line) 'pointer)
              ;; ...to update its children.
              change-children filter)

        ;; And then return the updated line.
        (update-line widget-line))
      
      ;; Otherwise, we will just change each property separately.
      (begin
        ;; Returns the field list, given the field name
        (letrec ((get-field (lambda (name)
                              (assq name (cdr widget-line)))))
          ;; Assign the widget pointer to a variable
          (let ((widget-pointer (cadr (get-field 'pointer))))
            ;;
            ;; And then we have (almost) no choice, because MrEd's function names differ,
            ;; and there is no real rule, therefore, we have to enumerate each case!
            ;; (We could, of course, make the code a bit shorter but it wouldn't be
            ;; worth doing it).
            ;;
            (case last-value-changed
              ;; These ones begin with "set"...
              ((label) 
               (send widget-pointer set-label (cadr (convert-elt (get-field 'label)))))
              ((selection) 
               (send widget-pointer set-selection (cadr (convert-elt (get-field 'selection)))))
              ((value) 
               (send widget-pointer set-value (cadr (convert-elt (get-field 'value)))))
              ((init-value) 
               (send widget-pointer set-value (cadr (convert-elt (get-field 'init-value)))))
              ((range) 
               (send widget-pointer set-range (cadr (convert-elt (get-field 'range)))))
              
              ;; These ones have no prefix...
              ((horiz-margin) 
               (send widget-pointer horiz-margin (cadr (convert-elt (get-field 'horiz-margin)))))
              ((vert-margin) 
               (send widget-pointer vert-margin (cadr (convert-elt (get-field 'vert-margin)))))
              ((min-width) 
               (send widget-pointer min-width (cadr (convert-elt (get-field 'min-width)))))
              ((min-height) 
               (send widget-pointer min-height (cadr (convert-elt (get-field 'min-height)))))
              ((stretchable-width) 
               (send widget-pointer 
                     stretchable-width 
                     (cadr (convert-elt (get-field 'stretchable-width)))))
              ((stretchable-height) 
               (send widget-pointer 
                     stretchable-height 
                     (cadr (convert-elt (get-field 'stretchable-height)))))
              ((border) 
               (send widget-pointer border (cadr (convert-elt (get-field 'border)))))
              ((spacing) 
               (send widget-pointer spacing (cadr (convert-elt (get-field 'spacing)))))
              
              ;; These ones are really weird !
              ((width)
               (send widget-pointer resize 
                     (cadr (convert-elt (get-field 'width))) 
                     (send widget-pointer get-height)))
              ((height) 
               (send widget-pointer resize 
                     (send widget-pointer get-width) 
                     (cadr (convert-elt (get-field 'height)))))
              ((x) 
               (send widget-pointer move 
                     (cadr (convert-elt (get-field 'x))) 
                     (send widget-pointer get-y)))
              ((y)
               (send widget-pointer move 
                     (send widget-pointer get-x) 
                     (cadr (convert-elt (get-field 'y)))))
              ((enabled) 
               (send widget-pointer enable (cadr (convert-elt (get-field 'enabled)))))
              ((alignment) 
               (send widget-pointer set-alignment 
                     (car (eval (cadr (convert-elt (get-field 'alignment)))))
                     (cadr (eval (cadr (convert-elt (get-field 'alignment))))))))
           
            ;; We warn the parent that the flow has been modified
            ;(let ((parent-pointer (project-data-get-parent widget-line)))
            ;  (if parent-pointer
            ;      (send (get-property-value parent-pointer 'pointer) container-flow-modified)))
            
            ;; Then we return the unchanged line...
            widget-line)))))

